// Handling diff or fc for regression-test.

//go:build !windows

package platform

// GetDiffRunArray gets the command for diffing actual/expected output in a
// regression test.
func GetDiffRunArray(filename1, filename2 string) []string {
	return []string{"diff", "-u", filename1, filename2}
}
